//----------------------------------------------------------------------------
//	SOURCE:	Map.c
//	AUTHOR:	Ian Taylor
//	NOTES:	Main source file of Map demo application
//----------------------------------------------------------------------------

#include "Map.h"
#include "MapProj.h"
#include "MapCodes.h"

// To switch on conversion options uncomment this line
#ifdef _DEBUG
#define MAP_CONVERSION
#endif

// Constants
#define MAP_WINDOW_SX_MIN	256
#define MAP_WINDOW_SY_MIN	192
#define MAP_FLAG_SX_MIN		240
#define MAP_FLAG_SY_MIN		120

// Map menu ids
enum CMD
{
	CMD_NONE = 0,
	CMD_BOUNDARIES,
	CMD_SAVE,
#ifdef MAP_CONVERSION
	CMD_TO_TEXT,
	CMD_TO_EMF,
	CMD_TO_SYNOPSIS,
	CMD_TO_FINAL,
#endif
	CMD_ABOUT,
	CMD_EXIT,
	CMD_STANDARD,
	CMD_SHOWALL,
	CMD_UNSELECT,
	CMD_FLAG,
	CMD_REPORT,
	CMD_PREVPROJECTION,
	CMD_NEXTPROJECTION
};

// Conversion builds only
#ifdef MAP_CONVERSION
	void ConvertToText(void);
	void ConvertToEMF(void);
	void ConvertToSynopsis(void);
	void ConvertToFinal(void);
#endif

// Globals
typedef struct _DIB
{
	SIZE original;
	RECT client;
	HBITMAP hbitmap;
	HDC hdc;
	HGDIOBJ hother;
	RECT rect;
	BOOL dirty;
}
DIB;
local DIB dib_map;
local DIB dib_grey;
local DIB dib_flag;
local HBRUSH hbrush_back;
local HBRUSH hbrush_sea;
local SIZE screen;
local struct _ZOOM
{
	float w, n, e, s, factor;
}
zoom;
local enum Projection current_projection;
local enum Lines
{
	Lines_None = 0, Lines_Grid, Lines_All
}
current_lines;
local int current_flag;

// The windows
local HWND hmap;
local HWND hflag;
local HWND hreport;
local HWND htooltip;

// The projected map
local MAPPOINT projected[MAX_POINTS];
local int points;
local int instruction[MAX_INSTRUCTIONS];
local int instructions;

// The colourising callback
typedef COLORREF (*MAPFUNC)(int area);
#define RGB_NULL 0xFFFFFFFF
#define RGB_SEA PALETTERGB(0x00, 0x00, 0xFF)
#define RGB_BLACK PALETTERGB(0x00, 0x00, 0x00)

// Transformed points
local POINT transformed[MAX_POINTS];

local void Transform(void)
{
	const float sx = dib_map.rect.right / (zoom.e - zoom.w);
	const float sy = dib_map.rect.bottom / (zoom.s - zoom.n);
	int i;

	// Transform the points
	for (i = 0; i < points; i++)
	{
		transformed[i].x = (LONG)((projected[i].x - zoom.w) * sx);
		transformed[i].y = (LONG)((projected[i].y - zoom.n) * sy);
	}
	return;
}

local void DrawTransformed(HDC hdc, MAPFUNC func, enum Lines lines)
{
	HBRUSH hlastBrush = NULL;
	HPEN hlastPen = NULL;
	COLORREF rgbLast = RGB_NULL;
	const POINT* pt = transformed;
	int last = -1;
	int i = 0;

	// Plot the transformed polygons
	while (i < instructions)
	{
		// Draw each poly
		int area = instruction[i++];
		int n = instruction[i++];
		if (n != 0)
		{
			if (area != last)
			{
				// Select the appropriate brush
				COLORREF rgb = func(area);
				if (rgb != rgbLast)
				{
					// New brush colour
					HBRUSH hbrush = OS(CreateSolidBrush)(rgb);
					OS(SelectObject)(hdc, hbrush);
					if (hlastBrush != NULL)
						OS(DeleteObject)(hlastBrush);
					hlastBrush = hbrush;
					if (lines != Lines_All)
					{
						// New pen colour
						HPEN hpen = OS(CreatePen)(PS_SOLID, 0, rgb);
						OS(SelectObject)(hdc, hpen);
						if (hlastPen != NULL)
							OS(DeleteObject)(hlastPen);
						hlastPen = hpen;
					}
					rgbLast = rgb;
				}
				last = area;
			}
			ASSERT((n >= 3) && (n <= 1000));
			VERIFY(OS(Polygon)(hdc, pt, n));
			pt += n;
		}
	}
	OS(SelectObject)(hdc, OS(GetStockObject)(NULL_BRUSH));
	OS(SelectObject)(hdc, OS(GetStockObject)(BLACK_PEN));
	if (hlastPen != NULL)
		OS(DeleteObject)(hlastPen);
	if (hlastBrush != NULL)
		OS(DeleteObject)(hlastBrush);
	if (lines == Lines_Grid)
	{
		// Now plot the edges of the grid
		pt = transformed;
		i = 0;
		while (i < instructions)
		{
			int area = instruction[i++];
			int n = instruction[i++];
			if ((area == 0) && (n > 0))
			{
				ASSERT(n == 91);
				VERIFY(OS(Polygon)(hdc, pt, n));
			}
			pt += n;
		}
	}
	return;
}

local void CreateDIB(DIB* dib, HWND hwnd)
{
	// Create the DIB
	HDC hdc = OS(GetDC)(hwnd);
	ASSERT(screen.cx > 0);
	ASSERT(screen.cy > 0);
	ASSERT(hdc != NULL);
	dib->hdc = OS(CreateCompatibleDC)(hdc);
	ASSERT(dib->hdc != NULL);
	dib->hbitmap = OS(CreateCompatibleBitmap)(hdc, screen.cx, screen.cy);
	ASSERT(dib->hbitmap != NULL);
	dib->hother = OS(SelectObject)(dib->hdc, dib->hbitmap);
	ASSERT(dib->hother != NULL);
	dib->rect.left = 0;
	dib->rect.top = 0;
	if (hwnd != NULL)
		VERIFY(OS(SetStretchBltMode)(hdc, STRETCH_HALFTONE));
	return;
}

local void ResizeDIB(DIB* dib, int width, int height)
{
	// Resize the DIB
	ASSERT(width > 0);
	ASSERT(height > 0);
	ASSERT(dib->rect.left == 0);
	ASSERT(dib->rect.top == 0);
	if (width > screen.cx)
	{
		// Go into stretch mode
		dib->rect.right = screen.cx;
	}
	else
	{
		// The existing bitmap is big enough (size for antialiasing)
		dib->rect.right = (screen.cx / width) * width;
		ASSERT(dib->rect.right >= width);
	}
	if (height > screen.cy)
	{
		// Go into stretch mode
		dib->rect.bottom = height;
	}
	else
	{
		// The existing bitmap is big enough (size for antialiasing)
		dib->rect.bottom = (screen.cy / height) * height;
		ASSERT(dib->rect.bottom >= height);
	}
	return;
}

local void DestroyDIB(DIB* dib)
{
	// Destroy the DIB
	if (dib->hbitmap != NULL)
	{
		ASSERT(dib->hdc != NULL);
		ASSERT(dib->hother != NULL);
		VERIFY(OS(SelectObject)(dib->hdc, dib->hother));
		VERIFY(OS(DeleteObject)(dib->hbitmap));
		VERIFY(OS(DeleteDC)(dib->hdc));
	}
	return;
}

local void UpdateFlagDIB(int country)
{
	// Update the picture in the flag DIB
	ASSERT(dib_flag.dirty);
	VERIFY(OS(FillRect)(dib_flag.hdc, &dib_flag.rect, hbrush_back));
	ASSERT(dib_flag.rect.left == 0);
	ASSERT(dib_flag.rect.top == 0);
	if (country > 0)
		FlagDraw(dib_flag.hdc, dib_flag.rect.right, dib_flag.rect.bottom, 2.0f, country);
	dib_flag.dirty = FALSE;
	return;
}

local void FlagSizing(WPARAM direction, LPRECT rect)
{
	// Work out the extra pixels added for the frame
	int ex = dib_flag.original.cx - MAP_FLAG_SX_MIN;
	int ey = dib_flag.original.cy - MAP_FLAG_SY_MIN;

	// Work out the requested client area size
	int sx = rect->right - rect->left - ex;
	int sy = rect->bottom - rect->top - ey;

	// Work out the corrected window sizes
	int cx = sy * 2 + ex;
	int cy = sx / 2 + ey;

	// Constrain the resizing to the correct aspect ratio
	switch (direction)
	{
		case WMSZ_BOTTOM:
		case WMSZ_TOP:
			rect->right = rect->left + cx;
			if (rect->right > screen.cx)
			{
				// Too far
				rect->right = screen.cx;
				FlagSizing(WMSZ_RIGHT, rect);
			}
			break;

		case WMSZ_BOTTOMLEFT:
		case WMSZ_BOTTOMRIGHT:
		case WMSZ_LEFT:
		case WMSZ_RIGHT:
			rect->bottom = rect->top + cy;
			if (rect->bottom > screen.cx)
			{
				// Too far
				rect->bottom = screen.cy;
				FlagSizing(WMSZ_BOTTOM, rect);
			}
			break;

		case WMSZ_TOPLEFT:
		case WMSZ_TOPRIGHT:
			rect->top = rect->bottom - cy;
			if (rect->top < 0)
			{
				// Too far
				rect->top = 0;
				FlagSizing(WMSZ_BOTTOM, rect);
			}
			break;

		default:
			UNREACHABLE();
			break;
	}
	return;
}

local LRESULT CALLBACK FlagWindowProc(HWND hwnd, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	// Handle window message for the flag window
	switch (umsg)
	{
		case WM_SIZING:
			// The user is sizing the window
			FlagSizing(wparam, (LPRECT)lparam);
			return TRUE;

		case WM_GETMINMAXINFO:
			// Don't let the window get too small
			{
				LPMINMAXINFO mmi = (LPMINMAXINFO)lparam;
				ASSERT(mmi != NULL);
				mmi->ptMinTrackSize.x = dib_flag.original.cx;
				mmi->ptMinTrackSize.y = dib_flag.original.cy;
			}
			return 0;

		case WM_SIZE:
			// The user has resized the window (always chain this message)
			if (wparam != SIZE_MINIMIZED)
			{
				VERIFY(OS(GetClientRect)(hwnd, &dib_flag.client));
				ResizeDIB(&dib_flag, LOWORD(lparam), HIWORD(lparam));
				dib_flag.dirty = TRUE;
				VERIFY(OS(InvalidateRect)(hwnd, NULL, FALSE));
			}
			break;

		case WM_PAINT:
			// Repaint the bitmap
			{
				PAINTSTRUCT ps;
				if (OS(BeginPaint)(hwnd, &ps))
				{
					if (dib_flag.dirty)
						UpdateFlagDIB(current_flag);
					ASSERT(dib_flag.client.left == 0);
					ASSERT(dib_flag.client.top == 0);
					VERIFY(OS(StretchBlt)(ps.hdc, 0, 0, dib_flag.client.right, dib_flag.client.bottom, dib_flag.hdc, 0, 0, dib_flag.rect.right, dib_flag.rect.bottom, SRCCOPY));
					VERIFY(OS(EndPaint)(hwnd, &ps));
				}
			}
			break;

		case WM_CLOSE:
			// Actually, just hide it
			VERIFY(OS(ShowWindow)(hwnd, SW_HIDE));
			return 0;
	}
	return OS(DefWindowProc)(hwnd, umsg, wparam, lparam);
}

local HWND FlagCreate(HWND hparent, HICON hicon)
{
	// Create the flag window
	static const TCHAR classname[] = _T("MapFlagWindow");
	HINSTANCE hinstance = (HINSTANCE)OS(GetWindowLong)(hparent, GWL_HINSTANCE);
	const int style = WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_THICKFRAME|WS_MAXIMIZEBOX;
	RECT client = { 0, 0, MAP_FLAG_SX_MIN, MAP_FLAG_SY_MIN };
	WNDCLASS wc;
	HWND hwnd;

	// Register the window class
	wc.style = CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
	wc.lpfnWndProc = FlagWindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hinstance;
	wc.hIcon = hicon;
	wc.hCursor = OS(LoadCursor)(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = classname;
	ASSERT(wc.hCursor != NULL);
	VERIFY(OS(RegisterClass)(&wc));

	// Create the window
	VERIFY(OS(AdjustWindowRectEx)(&client, style, FALSE, WS_EX_CLIENTEDGE));
	dib_flag.original.cx = client.right - client.left;
	dib_flag.original.cy = client.bottom - client.top;
	VERIFY(OS(GetWindowRect)(hparent, &client));
    hwnd = OS(CreateWindowEx)(WS_EX_CLIENTEDGE, classname, _T("Map - Flag"), style,
							  client.right - dib_flag.original.cx - MAP_FLAG_SX_MIN, client.bottom,
							  dib_flag.original.cx + MAP_FLAG_SX_MIN, dib_flag.original.cy + MAP_FLAG_SY_MIN,
							  hparent, NULL, hinstance, 0);
	ASSERT(hwnd != NULL);
	OS(ShowWindow)(hwnd, SW_NORMAL);
	return hwnd;
}

void MapRepaint(void)
{
	// Queue up a map redraw
	dib_map.dirty = TRUE;
	VERIFY(OS(InvalidateRect)(hmap, NULL, FALSE));
	return;
}

local void MakeMapDirty(void)
{
	// Force a redraw later on
	ASSERT(zoom.w >= 0);
	ASSERT(zoom.e >= zoom.w);
	ASSERT(zoom.e <= 360);
	ASSERT(zoom.n >= 0);
	ASSERT(zoom.s > zoom.n);
	ASSERT(zoom.s <= 180);
	Transform();
	VERIFY(OS(FillRect)(dib_grey.hdc, &dib_grey.rect, OS(GetStockObject)(BLACK_BRUSH)));
	DrawTransformed(dib_grey.hdc, ColourGrey, Lines_None);
	MapRepaint();
	return;
}

local void ResetProjectionEx(int proj, int scroll)
{
	MAPPROJECT p;
	int i = 0;
	int n;

	// Set the projection
	instructions = 0;
	p.projection = &projection[proj];
	p.rx = 0;
	p.ry = 0;
	p.dst = projected;
	p.src = raw_point;
	p.scroll = scroll;

	// Project the polygons
	while (i < raw_instructions)
		if (raw_instruction[i] < 0)
		{
			// Move to the next region
			if (++p.rx == 24)
			{
				p.rx = 0;
				p.ry++;
			}
			i++;
		}
		else
		{
			// This is a poly to add
			p.n = raw_instruction[i + 1];
			n = p.projection->projector(&p);
			ASSERT((n == 0) || (n >= 2));
			instruction[instructions++] = raw_instruction[i];
			instruction[instructions++] = n;
			p.dst += n;
			p.src += p.n;
			i += 2;
		}
	points = p.dst - projected;
	ASSERT((instructions > 0) && (instructions < dimof(instruction)));
	ASSERT((points > 0) && (points < dimof(projected)));
	return;
}

local void ResetProjection(int scroll)
{
	// Set the scroll bar
	OS(SetScrollPos)(hmap, SB_HORZ, scroll, TRUE);
	ResetProjectionEx(current_projection, scroll);
	MakeMapDirty();
	return;
}

local void ZoomTo(const MAPPOINT* map, float factor)
{
	// Zoom in/out to a particular point
	zoom.w = map->x - (map->x - zoom.w) * zoom.factor / factor;
	if (zoom.w < 0)
		zoom.w = 0;
	zoom.e = zoom.w + 360 / factor;
	if (zoom.e > 360)
	{
		zoom.e = 360;
		zoom.w = 360 - 360 / factor;
	}
	zoom.n = map->y - (map->y - zoom.n) * zoom.factor / factor;
	if (zoom.n < 0)
		zoom.n = 0;
	zoom.s = zoom.n + 180 / factor;
	if (zoom.s > 180)
	{
		zoom.s = 180;
		zoom.n = 180 - 180 / factor;
	}
	zoom.factor = factor;
	MakeMapDirty();
	return;
}

local void ZoomOut(void)
{
	// Zoom fully out
	zoom.w = 0;
	zoom.n = 0;
	zoom.e = 360;
	zoom.s = 180;
	zoom.factor = 1;
	return;
}

local void ProjectionName(TCHAR buffer[256], int proj, TCHAR separator)
{
	// Construct a full projection name
	const MAPPROJECTION* p = &projection[proj];
	TCHAR text[8];
	LPTSTR dst;
	int year;
	OS(lstrcat)(buffer, projection_name[proj][0]);
	text[0] = separator;
	text[1] = _T('(');
	text[2] = _T('\0');
	OS(lstrcat)(buffer, text);
	if (projection_name[proj][1] != NULL)
	{
		// Append the author
		OS(lstrcat)(buffer, projection_name[proj][1]);
		OS(lstrcat)(buffer, _T(", "));
	}
	if (p->year & 1)
		OS(lstrcat)(buffer, _T("c."));
	year = p->year >> 1;
	if (year < 0)
		year = -year;
	dst = text + dimof(text);
	*--dst = _T('\0');
	while (year > 0)
	{
		// Convert the year to denary
		*--dst = (TCHAR)(year % 10 + _T('0'));
		year /= 10;
	}
	OS(lstrcat)(buffer, dst);
	if (p->year < 0)
		OS(lstrcat)(buffer, _T("BC)"));
	else if (p->year < 2000)
		OS(lstrcat)(buffer, _T("AD)"));
	else
		OS(lstrcat)(buffer, _T(")"));
	return;
}

local void SaveProjection(void)
{
	OPENFILENAME of = { sizeof(of) };
	TCHAR path[2048] = _T("");
	TCHAR description[256] = _T("Map - chilliant.com - ");
	HENHMETAFILE emf;
	HDC hdc;
	int i;

	// Save the current projection to a file
	of.hwndOwner = hmap;
	of.lpstrFilter = _T("Windows Enhanced Metafile (*.emf)\0*.emf\0\0");
	of.lpstrDefExt = _T("emf");
	of.lpstrFile = path;
	of.nMaxFile = dimof(path) - 4;
	of.Flags = OFN_OVERWRITEPROMPT|OFN_PATHMUSTEXIST|OFN_HIDEREADONLY;
	if (OS(GetSaveFileName)(&of) == 0)
		return;

	// Create a metafile with an appropriate description
	ProjectionName(description, current_projection, _T(' '));
	hdc = OS(CreateEnhMetaFile)(NULL, path, NULL, description);
	ASSERT(hdc != NULL);

	// Plot the coordinates in a canvas no bigger than 32767 units
	for (i = 0; i < points; i++)
	{
		transformed[i].x = (LONG)(projected[i].x * 91);
		transformed[i].y = (LONG)(projected[i].y * 91);
	}
	DrawTransformed(hdc, ColourDefault, current_lines);
	Transform();

	// Close the metafile
	emf = OS(CloseEnhMetaFile)(hdc);
	ASSERT(emf != NULL);
	VERIFY(OS(DeleteEnhMetaFile)(emf));
	return;
}

local void SetProjection(enum Projection proj, int scroll)
{
	TCHAR caption[256];

	// Set the scrollbar and projection
	ASSERT((proj >= 0) && (proj < Projection_Count));
	ZoomOut();
	current_projection = proj;
	ResetProjection(scroll);
	OS(lstrcpy)(caption, _T("Map - "));
	ProjectionName(caption, proj, _T(' '));
	OS(SetWindowText)(hmap, caption);
	return;
}

local void UpdateMapDIB(void)
{
	// Update the picture in the map DIB
	ASSERT(dib_map.dirty);
	VERIFY(OS(FillRect)(dib_map.hdc, &dib_map.rect, hbrush_back));
	DrawTransformed(dib_map.hdc, ColourDefault, current_lines);
	dib_map.dirty = FALSE;
	return;
}

local void OnSwitch(HWND hchild)
{
	// Switch to another window
	ASSERT(hchild != NULL);
	OS(ShowWindow)(hchild, OS(IsWindowVisible)(hchild) ? SW_HIDE : SW_SHOW);
	return;
}

local void OnZoom(int mx, int my, BOOL out)
{
	float factor;
	MAPPOINT map;

	// Zoom in or out
	ASSERT(dib_map.client.left == 0);
	ASSERT(dib_map.client.top == 0);
	map.x = mx * (zoom.e - zoom.w) / dib_map.client.right + zoom.w;
	map.y = my * (zoom.s - zoom.n) / dib_map.client.bottom + zoom.n;
	if (!out)
	{
		// Zoom in
		if (zoom.factor >= 200)
			return;
		factor = zoom.factor * 1.1F;
	}
	else if (zoom.factor < 1.1F)
	{
		// Zoom fully out
		factor = 1;
	}
	else
	{
		// Zoom partially out
		factor = zoom.factor / 1.1F;
	}
	ZoomTo(&map, factor);
	return;
}

local void OnScroll(int dir)
{
	// Scroll to the appropriate place
	int scroll = (OS(GetScrollPos)(hmap, SB_HORZ) + dir + 24) % 24;
	ResetProjection(scroll);
	return;
}

local void OnSelect(int region)
{
	// Select a particular region
	int country = RegionToCountry(region);
	FlagSelect(country);
	ReportSelect(country);
	return;
}

local void OnCommand(enum CMD cmd)
{
	static const TCHAR about[] = _T("Copyright  2003-6 Ian Taylor\r\n\r\nSee www.chilliant.com for more details");

	// A menu item has been selected
	switch (cmd)
	{
		case CMD_BOUNDARIES:
			// Toggle boundaries
			current_lines = (current_lines == Lines_Grid) ? Lines_All : Lines_Grid;
			MapRepaint();
			break;

		case CMD_SAVE:
			// Save the current projection
			SaveProjection();
			break;

		case CMD_PREVPROJECTION:
			// Go to the previous projection
			SetProjection((current_projection + Projection_Count - 1) % Projection_Count, 0);
			break;

		case CMD_NEXTPROJECTION:
			// Go to the next projection
			SetProjection((current_projection + 1) % Projection_Count, 0);
			break;

		case CMD_STANDARD:
			// Original size
			VERIFY(OS(ShowWindow)(hmap, SW_RESTORE));
			VERIFY(OS(SetWindowPos)(hmap, NULL, 0, 0, dib_map.original.cx, dib_map.original.cy, SWP_NOMOVE|SWP_NOZORDER));
			break;

		case CMD_SHOWALL:
			// Show all
			ZoomOut();
			MakeMapDirty();
			break;

		case CMD_UNSELECT:
			// About
			OnSelect(0);
			break;

		case CMD_FLAG:
			// Flag window
			OnSwitch(hflag);
			break;

		case CMD_REPORT:
			// Report window
			OnSwitch(hreport);
			break;

		case CMD_ABOUT:
			// Unselect
			OS(MessageBox)(hmap, about, _T("About Map"), MB_ICONINFORMATION);
			break;

		case CMD_EXIT:
			// Exit
			OS(PostQuitMessage)(0);
			break;

#ifdef MAP_CONVERSION
		case CMD_TO_TEXT:
			// Convert EMF to text
			ConvertToText();
			break;

		case CMD_TO_EMF:
			// Convert text to EMF
			ConvertToEMF();
			break;

		case CMD_TO_SYNOPSIS:
			// Convert text to a synopsis
			ConvertToSynopsis();
			break;

		case CMD_TO_FINAL:
			// Convert text to the final format
			ConvertToFinal();
			break;
#endif

		default:
			// Probably a projection
			if (cmd >= 0x100)
				SetProjection(cmd & 0xFF, 0);
			break;
	}
	return;
}

local int AreaFromLPARAM(LPARAM lparam)
{
	// Convert a client point to a region index
	int x = (int)((double)LOWORD(lparam) / dib_map.client.right * dib_grey.rect.right);
	int y = (int)((double)HIWORD(lparam) / dib_map.client.bottom * dib_grey.rect.bottom);
	COLORREF rgb = OS(GetPixel)(dib_grey.hdc, x, y);
	ASSERT(dib_map.client.left == 0);
	ASSERT(dib_map.client.top == 0);
	ASSERT(dib_grey.client.left == 0);
	ASSERT(dib_grey.client.top == 0);
	ASSERT(GetRValue(rgb) == GetGValue(rgb));
	ASSERT(GetRValue(rgb) == GetBValue(rgb));
	return GetRValue(rgb);
}

local LRESULT CALLBACK MapWindowProc(HWND hwnd, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	switch (umsg)
	{
		case WM_MOUSEMOVE:
			// Track the tooltip
			if (htooltip != NULL)
			{
				POINT pt;
				int new_region;
				pt.x = LOWORD(lparam);
				pt.y = HIWORD(lparam);
				new_region = AreaFromLPARAM(lparam);
				if (!OS(ClientToScreen)(hwnd, &pt))
					new_region = -1;
				TooltipTrack(&pt, new_region);
			}
			break;

		case WM_MOUSELEAVE:
			// The mouse has left the window
			if (htooltip != NULL)
				TooltipTrack(NULL, -1);
			break;

		case WM_GETMINMAXINFO:
			// Don't let the window get too small
			{
				LPMINMAXINFO mmi = (LPMINMAXINFO)lparam;
				ASSERT(mmi != NULL);
				mmi->ptMinTrackSize.x = MAP_WINDOW_SX_MIN;
				mmi->ptMinTrackSize.y = MAP_WINDOW_SY_MIN;
			}
			return 0;

		case WM_SIZE:
			// The user has resized the window (always chain this message)
			if (wparam != SIZE_MINIMIZED)
			{
				VERIFY(OS(GetClientRect)(hwnd, &dib_map.client));
				dib_grey.client = dib_map.client;
				ResizeDIB(&dib_map, LOWORD(lparam), HIWORD(lparam));
				ResizeDIB(&dib_grey, LOWORD(lparam), HIWORD(lparam));
				MakeMapDirty();
			}
			break;

		case WM_LBUTTONDOWN:
			// Select current
			OnSelect(AreaFromLPARAM(lparam));
			return 0;

		case WM_MBUTTONDOWN:
			// Previous projection
			OnCommand(CMD_PREVPROJECTION);
			return 0;

		case WM_RBUTTONDOWN:
			// Next projection
			OnCommand(CMD_NEXTPROJECTION);
			return 0;

		case WM_KEYDOWN:
			// Trap arrow keys
			switch (wparam)
			{
				case VK_UP:
					// Up
					OnCommand(CMD_PREVPROJECTION);
					return 0;

				case VK_DOWN:
					// Down
					OnCommand(CMD_NEXTPROJECTION);
					return 0;

				case VK_LEFT:
					// Scroll left
					OnScroll(-1);
					return 0;

				case VK_RIGHT:
					// Scroll right
					OnScroll(1);
					return 0;
			}
			break;

		case WM_MOUSEWHEEL:
			// Zoom or scroll left or right
			if (LOWORD(wparam) & MK_CONTROL)
			{
				// Scroll
				OnScroll((HIWORD(wparam) & 0x8000) ? -1 : 1);
			}
			else
			{
				// This is a zoom request
				POINT top_left;
				top_left.x = LOWORD(lparam);
				top_left.y = HIWORD(lparam);
				OS(ScreenToClient)(hwnd, &top_left);
				OnZoom(top_left.x, top_left.y, HIWORD(wparam) & 0x8000);
			}
			return 0;

		case WM_LBUTTONDBLCLK:
			// Zoom current
			RegionZoom(AreaFromLPARAM(lparam));
			return 0;


		case WM_COMMAND:
			// The user has selected a menu item
			OnCommand(LOWORD(wparam));
			return 0;

		case WM_HSCROLL:
			// The bottom scrollbar has changed
			switch (LOWORD(wparam))
			{
				case SB_LINELEFT:
				case SB_PAGELEFT:
					// Scroll left
					OnScroll(-1);
					return 0;

				case SB_LINERIGHT:
				case SB_PAGERIGHT:
					// Scroll right
					OnScroll(1);
					return 0;

				case SB_THUMBTRACK:
					// Dragging
					ResetProjection(HIWORD(wparam));
					return 0;

				case SB_THUMBPOSITION:
					// Dragged
					ResetProjection(HIWORD(wparam));
					return 0;
			}
			break;

		case WM_PAINT:
			// Repaint the bitmap
			{
				PAINTSTRUCT ps;
				if (OS(BeginPaint)(hwnd, &ps))
				{
					if (dib_map.dirty)
						UpdateMapDIB();
					ASSERT(dib_map.client.left == 0);
					ASSERT(dib_map.client.top == 0);
					VERIFY(OS(StretchBlt)(ps.hdc, 0, 0, dib_map.client.right, dib_map.client.bottom, dib_map.hdc, 0, 0, dib_map.rect.right, dib_map.rect.bottom, SRCCOPY));
					VERIFY(OS(EndPaint)(hwnd, &ps));
				}
			}
			break;

		case WM_CLOSE:
			// Explicit close
			OS(PostQuitMessage)(0);
			return 0;
	}
	return OS(DefWindowProc)(hwnd, umsg, wparam, lparam);
}

local HMENU MakeMenu(void)
{
#define MENU_RAW(text)	_T(text) _T("\0")
#define MENU_ADD(text)	_T("+") MENU_RAW(text)
#define MENU_ITEM(text)	MENU_RAW(text)
#define MENU_SEPARATOR	_T("-")
#define MENU_END		_T(".")
	LPCTSTR menu =
		MENU_ADD("&Map")
			MENU_ITEM("Show &Boundaries\tCtrl+B")
			MENU_SEPARATOR
			MENU_ITEM("&Save As...\tCtrl+S")
#ifdef MAP_CONVERSION
			MENU_ITEM("Clipboard To MAP.&TXT")
			MENU_ITEM("MAP.TXT to &Clipboard")
			MENU_ITEM("MAP.TXT to S&ynopsis")
			MENU_ITEM("MAP.TXT to &MapData.h")
#endif
			MENU_SEPARATOR
			MENU_ITEM("&About...\tF1")
			MENU_ITEM("E&xit\tAlt+F4")
		MENU_ADD("&Window")
			MENU_ITEM("&Original Size\tCtrl+O")
			MENU_ITEM("Show &All\tCtrl+A")
			MENU_ITEM("Select &None\tEsc")
			MENU_SEPARATOR
			MENU_ITEM("&Flags\tCtrl+F")
			MENU_ITEM("&Datasheet\tCtrl+D")
		MENU_ADD("&Projection")
			MENU_ITEM("&Previous Projection\tUp")
			MENU_ITEM("&Next Projection\tDown")
			MENU_SEPARATOR
		MENU_END

		MENU_RAW("&Cylindrical")
		MENU_RAW("Co&nic")
		MENU_RAW("&Azimuthal")
		MENU_RAW("&Pseudocylindrical")
		MENU_RAW("&Other");
	TCHAR text[256];
	HMENU hmain, hmenu = NULL, hsub = NULL;
	int group = 0;
	int subid = 0;
	int i = 0;

	// Create a menu
	hmain = OS(CreateMenu)();
	ASSERT(hmain != NULL);
	while (*menu != '.')
		switch (*menu)
		{
			case '+':
				// Start a new submenu
				hmenu = OS(CreatePopupMenu)();
				ASSERT(hmenu != NULL);
				VERIFY(OS(AppendMenu)(hmain, MF_POPUP, (UINT)hmenu, ++menu));
				while (*(menu++) != '\0');
				break;

			case '-':
				// Separator
				VERIFY(OS(AppendMenu)(hmenu, MF_SEPARATOR, 0, NULL));
				menu++;
				break;

			default:
				// Menu item
				VERIFY(OS(AppendMenu)(hmenu, MF_STRING, ++i, menu));
				while (*(menu++) != '\0');
				break;
		}
	menu++;
	for (i = 0; i < Projection_Count; i++)
	{
		UINT flags = MF_STRING;
		switch (projection[i].group - group)
		{
			case 0:
				// Same subgroup
				break;
			case 1:
				// Next subgroup
				VERIFY(OS(AppendMenu)(hsub, MF_SEPARATOR, 0, NULL));
				break;
			case 2:
				// Next column
				flags = MF_STRING|MF_MENUBARBREAK;
				break;
			default:
				// Start a new submenu
				subid = projection[i].group >> 4;
				hsub = OS(CreatePopupMenu)();
				ASSERT(hsub != NULL);
				VERIFY(OS(AppendMenu)(hmenu, MF_POPUP, (UINT)hsub, menu));
				while (*(menu++) != '\0');
				subid <<= 8;
				break;
		}
		group = projection[i].group;
		text[0] = _T('\0');
		ProjectionName(text, i, _T('\t'));
		VERIFY(OS(AppendMenu)(hsub, flags, i + subid, text));
	}
	return hmain;
}

local COLORREF ColourIcon(int area)
{
	// Draw areas with their region colours
	return (area == 0) ? RGB(0,0,255) : RGB(0,255,0);
}

local COLORREF ColourIconMask(int area)
{
	// Draw areas with their region colours
	USE(area);
	return RGB(0,0,0);
}

local HICON MakeIcon(void)
{
	HDC shdc = OS(GetDC)(NULL);
	HDC hdc = OS(CreateCompatibleDC)(shdc);
	HBITMAP hbitmap = OS(CreateCompatibleBitmap)(shdc, 256, 256);
	HBITMAP hmask = OS(CreateBitmap)(256, 256, 1, 1, NULL);
	HGDIOBJ hprev = OS(SelectObject)(hdc, hmask);
	RECT rect = { 0, 0, 256, 256 };
	ICONINFO iconinfo;
	HICON hicon;
	int i;

	// Create an icon from an interesting map projection
	ResetProjectionEx(Projection_ObliqueOrthographic0, 0);
	for (i = 0; i < points; i++)
	{
		transformed[i].x = (LONG)((projected[i].x - 90) * 255 / 180);
		transformed[i].y = (LONG)(projected[i].y * 255 / 180);
	}

	// Draw the mask
	VERIFY(OS(FillRect)(hdc, &rect, OS(GetStockObject)(WHITE_BRUSH)));
	DrawTransformed(hdc, ColourIconMask, Lines_None);
	OS(SelectObject)(hdc, hbitmap);

	// Draw the colours
	VERIFY(OS(FillRect)(hdc, &rect, OS(GetStockObject)(BLACK_BRUSH)));
	DrawTransformed(hdc, ColourIcon, Lines_None);
	OS(SelectObject)(hdc, hprev);

	// Construct the icon
	iconinfo.fIcon = TRUE;
	iconinfo.hbmMask = hmask;
	iconinfo.hbmColor = hbitmap;
	hicon = OS(CreateIconIndirect)(&iconinfo);
	VERIFY(OS(DeleteObject)(hbitmap));
	VERIFY(OS(DeleteObject)(hmask));
	VERIFY(OS(DeleteDC)(hdc));
	VERIFY(OS(ReleaseDC)(NULL, shdc));
	return hicon;
}

local HACCEL MakeAccel(void)
{
	// Create a keyboard accelerator table
	static const ACCEL accel[] =
	{
		{ FVIRTKEY|FCONTROL, 'A', CMD_SHOWALL },
		{ FVIRTKEY|FCONTROL, 'B', CMD_BOUNDARIES },
		{ FVIRTKEY|FCONTROL, 'D', CMD_REPORT },
		{ FVIRTKEY|FCONTROL, 'F', CMD_FLAG },
		{ FVIRTKEY|FCONTROL, 'O', CMD_STANDARD },
		{ FVIRTKEY|FCONTROL, 'S', CMD_SAVE },
		{ FVIRTKEY, VK_ESCAPE, CMD_UNSELECT },
		{ FVIRTKEY, VK_F1, CMD_ABOUT }
	};
	return OS(CreateAcceleratorTable)((LPACCEL)accel, dimof(accel));
}

int CALLBACK WinMain(HINSTANCE hinstance, HINSTANCE hprevious, LPSTR cmdline, int cmdshow)
{
	static const TCHAR classname[] = _T("MapWindow");
	const int style = WS_OVERLAPPEDWINDOW|WS_HSCROLL;
	WNDCLASS wc;
	HMENU hmenu;
	HICON hicon;
	HACCEL haccel;
	MSG msg;

	// Initialisation
#ifdef MAPDLL_HASH
	DLLLoad();
#endif
	hinstance = OS(GetModuleHandle)(NULL);
	Decompress();
	FlagDecompress();
	ProjectionDecompress();
	hmenu = MakeMenu();
	hicon = MakeIcon();
	haccel = MakeAccel();

	// Unused parameters
	USE(hprevious);
	USE(cmdline);

	// Register the window class
	wc.style = CS_HREDRAW|CS_VREDRAW|CS_OWNDC|CS_DBLCLKS;
	wc.lpfnWndProc = MapWindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hinstance;
	wc.hIcon = hicon;
	wc.hCursor = OS(LoadCursor)(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = classname;
	ASSERT(wc.hCursor != NULL);
	VERIFY(OS(RegisterClass)(&wc));

	// Calculate the standard size
	current_lines = Lines_Grid;
	current_flag = 0;
	screen.cx = OS(GetSystemMetrics)(SM_CXSCREEN);
	screen.cy = OS(GetSystemMetrics)(SM_CYSCREEN);
	dib_map.client.left = 0;
	dib_map.client.top = 0;
	dib_map.client.right = screen.cx / 2;
	dib_map.client.bottom = screen.cx / 4;
	VERIFY(OS(AdjustWindowRectEx)(&dib_map.client, style, TRUE, WS_EX_CLIENTEDGE));
	dib_map.original.cx = dib_map.client.right - dib_map.client.left;
	dib_map.original.cy = dib_map.client.bottom - dib_map.client.top;
	dib_grey = dib_map;

	// Create the grey-scale dib for looking up mouse points
	CreateDIB(&dib_grey, NULL);
	ResizeDIB(&dib_grey, screen.cx / 2, screen.cx / 4);

	// Create the map window
	hmap = OS(CreateWindowEx)(WS_EX_CLIENTEDGE, classname, _T("Map by Ian Taylor"), style, 0, 0,
							  dib_map.original.cx, dib_map.original.cy, HWND_DESKTOP, hmenu, hinstance, 0);
	ASSERT(hmap != NULL);
	VERIFY(OS(SetScrollRange)(hmap, SB_HORZ, 0, 23, TRUE));
	CreateDIB(&dib_map, hmap);
	ResizeDIB(&dib_map, screen.cx / 2, screen.cx / 4);
	hbrush_back = OS(CreateSolidBrush)(PALETTERGB(0x00, 0x80, 0x80));
	ASSERT(hbrush_back != NULL);
	hbrush_sea = OS(CreateSolidBrush)(RGB_SEA);
	ASSERT(hbrush_sea != NULL);

	// Create our child windows
	htooltip = TooltipCreate(hmap);
#ifdef MAP_CONVERSION
	hreport = NULL;
	hflag = NULL;
#else
	hreport = ReportCreate(hmap, hicon);
	hflag = FlagCreate(hmap, hicon);
	ASSERT(hflag != NULL);
	CreateDIB(&dib_flag, hflag);
	ResizeDIB(&dib_flag, screen.cx / 2, screen.cx / 4);
	FlagSelect(0);
#endif

	// Display the map
	current_projection = Projection_Trapizoidal0;
	ZoomOut();
	ResetProjection(0);
	MapRecolour();
	OS(ShowWindow)(hmap, cmdshow);

	// Main message pump
	msg.message = WM_NULL;
	while (OS(GetMessage)(&msg, NULL, 0, 0) == TRUE)
		if (!OS(TranslateAccelerator)(hmap, haccel, &msg))
		{
			OS(TranslateMessage)(&msg);
			OS(DispatchMessage)(&msg);
		}
	ASSERT(msg.message == WM_QUIT);

	// Clean-up
	DestroyDIB(&dib_map);
	DestroyDIB(&dib_grey);
#ifndef MAP_CONVERSION
	DestroyDIB(&dib_flag);
#endif
	return msg.wParam;
}

void FlagSelect(int country)
{
	// Select a flag to display
	if (current_flag != country)
	{
		current_flag = country;
		dib_flag.dirty = TRUE;
		VERIFY(OS(InvalidateRect)(hflag, NULL, FALSE));
	}
	return;
}

local BOOL RegionZoomEx(int region)
{
	// Scan for a region to zoom in to
	const float expansion = 2.f;
	const MAPPOINT* pt = projected;
	float min_x = 999;
	float max_x = -999;
	float min_y = 999;
	float max_y = -999;
	float zx, zy, zz;
	int i = 0;
	int j;
	BOOL consider;

	// Look for the bounding rectangle
	while (i < instructions)
	{
		// Draw each poly
		int area = instruction[i++];
		int n = instruction[i++];
		if (n == 0)
			return FALSE;
		j = raw_area[area].code;
		consider = (j == region);
		if (!consider && (RegionToCountry(j) == region))
		{
			// Work out if we should consider regions "close" to the current extent
			const float range = 2;
			if ((max_x <= min_x) || (max_y <= min_y) || ((pt->x > (min_x - range)) && (pt->x < (max_x + range))) || ((pt->y > (min_y - range)) && (pt->y < (max_y + range))))
				consider = TRUE;
		}
		if (consider)
		{
			// This polygon is considered part of the region
			for (j = 0; j < n; j++)
			{
				if (pt[j].x < min_x)
					min_x = pt[j].x;
				if (pt[j].x > max_x)
					max_x = pt[j].x;
				if (pt[j].y < min_y)
					min_y = pt[j].y;
				if (pt[j].y > max_y)
					max_y = pt[j].y;
			}
		}
		pt += n;
	}
	zx = max_x - min_x;
	zy = max_y - min_y;
	if ((zx <= 0) || (zy <= 0))
		return FALSE;
	if (zx < (zy * 2))
	{
		// Narrow region
		zx = zy * 2;
		min_x = (min_x + max_x - zx) / 2;
		max_x = min_x + zx;
	}
	else
	{
		// Fat region
		zy = zx / 2;
		min_y = (min_y + max_y - zy) / 2;
		max_y = min_y + zy;
	}
	zz = (expansion - 1) / 2;
	min_x -= zx * zz;
	max_x += zx * zz;
	min_y -= zy * zz;
	max_y += zy * zz;
	if (min_x < 0)
	{
		min_x = 0;
		max_x = zx * expansion;
		if (max_x > 360)
			return FALSE;
	}
	if (max_x > 360)
	{
		max_x = 360;
		min_x = 360 - zx * expansion;
		if (min_x < 0)
			return FALSE;
	}
	if (min_y < 0)
	{
		min_y = 0;
		max_y = zy * expansion;
		if (max_y > 180)
			return FALSE;
	}
	if (max_y > 180)
	{
		max_y = 180;
		min_y = 180 - zy * expansion;
		if (min_y < 0)
			return FALSE;
	}
	zoom.w = min_x;
	zoom.n = min_y;
	zoom.e = max_x;
	zoom.s = max_y;
	zoom.factor = 360 / (max_x - min_x);
	MakeMapDirty();
	return TRUE;
}

void RegionZoom(int region)
{
	// Zoom in on a particular country
	if (region == 0)
	{
		// Double clicking the ocean shows the whole map
		OnCommand(CMD_SHOWALL);
		return;
	}
	if (!RegionZoomEx(region))
	{
		// Need to reset the project projection and try again
		switch (region)
		{
			case Code_AQ:
				// Antartical prefers a polar projection
				SetProjection(Projection_PolarOrthographic0, 0);
				break;
			case Code_RU:
				// Russia prefers a conic projection
				SetProjection(Projection_Polyconic0, 18);
				break;
			default:
				// Equirectangular
				SetProjection(Projection_Trapizoidal0, 0);
				break;
		}
		RegionZoomEx(region);
	}
	return;
}
